// CCNMediaTerm.cpp : Implementation of CCCNMediaTerm

#include "stdafx.h"
#include "stdlib.h"

#include "CCNSMT.h"
#include "CCNMediaTerm.h"
#include "AudioCodecs.h"

extern "C" {
#include "../MTC/DSP/FIRCoefficients.h"
}

#include <string>
#include <vector>

using namespace std;

long nextCookie;

// #define NO_RECV_MIXER

int Connect(AudioSource *source, AudioSink *sink)
{
	source->SubscribeAudioSink(sink);
	sink->SubscribeAudioSource(source);
	return 0;
}

int Disconnect(AudioSource *source, AudioSink *sink)
{
	source->UnsubscribeAudioSink(sink);
	sink->UnsubscribeAudioSource(source);
	return 0;
}

int
CCCNMediaTerm::SetTraceLevel() {
	long SystemMask = 0;
	GetRegKeyLong(HKEY_CURRENT_USER, "Software\\Cisco Systems\\MTC\\Tracing", "MTC", 0x100000);
	if ((SystemMask = GetRegKeyLong(HKEY_CURRENT_USER, "Software\\Cisco Systems\\MTC\\Tracing", "AllComponents", 0x0)) == 0)
		SystemMask = GetRegKeyLong(HKEY_CURRENT_USER, "Software\\Cisco Systems\\MTC\\Tracing", "MTC", 0x100000);
	tracer.SetSystemMask(SystemMask);
	return NOERROR;
}


// #define WM_GRAPHNOTIFY WM_USER + 13


// using namespace std;

/////////////////////////////////////////////////////////////////////////////
// CCCNMediaTerm

HRESULT 
CCCNMediaTerm::Initialize()
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState())
	CSingleLock cLock(&Mutex);
	cLock.Lock();
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "Initialize\n");
	if (!initialized)
	{
		InitializeCriticalSection(&filesPlayingMutex);
		waveSource = new WaveAudioSource();
		waveSink = new WaveAudioSink();
		filePlayWaveSink = new WaveAudioSink();
		rtpSink = new RTPAudioSink();
		rtpSource = new RTPAudioSource();
		inputFIRTransformer = new FIRTransformer();
		inputFIRTransformer->SetCoefficients(FIR_TELECASTER_TX, FIR_TELECASTER_TX_NUM, 16);
		outputFIRTransformer = new FIRTransformer();
		outputFIRTransformer->SetCoefficients(FIR_TELECASTER_RX, FIR_TELECASTER_RX_NUM, 16);
		pcm2G711Transformer = new PCM2G711Transformer();
		g7112PcmTransformer = new G7112PCMTransformer();
		limiter = new Limiter(-8.0, 0.075, -0.00075);
		voiceActivityDetector = new VoiceActivityDetector();
		rtpJitterBuffer = new RTPJitterBuffer();
		fixedSizeAudioBuffer = new FixedSizeAudioBuffer();
		rxPCMMixer = new PCMMixer();
		txPCMMixer = new PCMMixer();
		rxTransformer = g7112PcmTransformer;
		txTransformer = pcm2G711Transformer;
		txDuration = 20;
		rxDuration = 20;
		filesPlaying.clear();
		extraFilePlay = NULL;
		brxRunning = false;
		btxRunning = false;
		bFilePlayRunning = false;
		nextCookie = 0;
		waveinDeviceID = WAVE_MAPPER;
		waveoutDeviceID = WAVE_MAPPER;
		initialized = true;
	}
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "~Initialize\n");
	return NOERROR;
}


HRESULT
CCCNMediaTerm::UnInitialize() 
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState())
	CSingleLock cLock(&Mutex);
	cLock.Lock();
	if (initialized)
	{
		StopRX();
		StopTX();
		RemoveNonPlayingFiles();
		if (g7112PcmTransformer)
		{
			delete g7112PcmTransformer;
		}
		if (pcm2G711Transformer)
		{
			delete pcm2G711Transformer;
		}
		if (waveSink)
		{
			delete waveSink;
		}
		if (filePlayWaveSink)
		{
			delete filePlayWaveSink;
		}
		if (waveSource)
		{
			delete waveSource;
		}
		if (rtpSource)
		{
			delete rtpSource;
		}
		if (rtpSink)
		{
			delete rtpSink;
		}
		if (inputFIRTransformer)
		{
			delete inputFIRTransformer;
		}
		if (outputFIRTransformer)
		{
			delete outputFIRTransformer;
		}
		if (limiter)
		{
			delete limiter;
		}
		if (voiceActivityDetector)
		{
			delete voiceActivityDetector;
		}
		if (rxPCMMixer)
		{
			delete rxPCMMixer;
		}
		if (txPCMMixer)
		{
			delete txPCMMixer;
		}
		if (rtpJitterBuffer)
		{
			delete rtpJitterBuffer;
		}
		if (fixedSizeAudioBuffer)
		{
			delete fixedSizeAudioBuffer;
		}
		DeleteCriticalSection(&filesPlayingMutex);
		initialized = false;
	}
	return NOERROR;
}

STDMETHODIMP 
CCCNMediaTerm::SetAudioCodecRX(long CompressionType, long MillisecPacketSize, long EchoCancellationValue, long G723BitRate)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState())
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "SetAudioCodecRX : %d %d %d %d\n", CompressionType, MillisecPacketSize, EchoCancellationValue, G723BitRate);
	CSingleLock cLock(&Mutex);
	cLock.Lock();
	SetTraceLevel();

	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "SetAudioCodecRX : brxRunning = %d\n", brxRunning);
	bool restart = false;
	int result(0);
	int resultCode(0);

	if (brxRunning)
	{
		result = StopRX();
		if (result < 0)
		{
			resultCode = -10;
		}
		restart = true;
	}

	if (resultCode == 0)
	{
		rxDuration = MillisecPacketSize;
		waveSink->SetFrameSize(16 * MillisecPacketSize);

		switch (CompressionType) {
		case Media_Payload_G711Ulaw64k:
			rxTransformer = g7112PcmTransformer;
			g7112PcmTransformer->SetULaw();
			break;
		case Media_Payload_G711Alaw64k: 
			rxTransformer = g7112PcmTransformer;
			g7112PcmTransformer->SetALaw();
			break;
		default:
			tracer.tracef(SDI_LEVEL_ERROR, "SetAudioCodecRX : Invalid Codec # %d\n", CompressionType);
			return -30;
		}

		if (restart)
		{
			result = StartRX(waveoutDeviceID);
			if (result < 0)
			{
				resultCode = -20;
			}
		}
	}

	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "~SetAudioCodecRX : %d\n", resultCode);
	return resultCode;
}



STDMETHODIMP 
CCCNMediaTerm::SetAudioCodecTX(long CompressionType, long MillisecPacketSize, long PrecedenceValue, long SilenceSuppression, unsigned short MaxFramesPerPacket, long G723BitRate)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState())
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "SetAudioCodecTX : %d %d %d %d %d %d\n", CompressionType, MillisecPacketSize, PrecedenceValue, SilenceSuppression, MaxFramesPerPacket, G723BitRate);
	CSingleLock cLock(&Mutex);
	cLock.Lock();
	SetTraceLevel();

	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "SetAudioCodecTX : btxRunning = %d\n", btxRunning);
	bool restart = false;
	int result(0);
	int resultCode(0);

	if (btxRunning)
	{
		result = StopTX();
		if (result < 0)
		{
			resultCode = -10;
		}
		restart = true;
	}

	// update btxVAD after StopTX, so that StopTX knows whether VAD was on or off the last time
	btxVAD = SilenceSuppression == 0 ? false : true;

	if (resultCode == 0)
	{
		txDuration = MillisecPacketSize;
		waveSource->SetBufferSize(MillisecPacketSize * 16);
		switch (CompressionType) 
		{
		case Media_Payload_G711Ulaw64k:
			txTransformer = pcm2G711Transformer;
			pcm2G711Transformer->SetULaw();
			break;
		case Media_Payload_G711Alaw64k:
			txTransformer = pcm2G711Transformer;
			pcm2G711Transformer->SetALaw();
			break;
		default:
			tracer.tracef(SDI_LEVEL_ERROR, "SetAudioCodecTX : Invalid Codec # %d\n", CompressionType);
			return -30;
		}

		if (restart)
		{
			result = StartTX(waveinDeviceID);
			if (result < 0)
			{
				resultCode = -20;
			}
		}
	}

	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "~SetAudioCodecTX : %d\n", resultCode);
	return resultCode;
}



STDMETHODIMP 
CCCNMediaTerm::SetAudioDestination(BSTR strHostName, long nUDPPortNumber)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState())
	USES_CONVERSION;
	LPSTR hostname = W2A((wchar_t *) strHostName);
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "SetAudioDestination %s %d\n", hostname, nUDPPortNumber);
	CSingleLock cLock(&Mutex);
	cLock.Lock();
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "SetAudioDestination : entered Mutex\n");
	int resultCode(0);

	bool bUseFixedTransmitPort = GetRegKeyBool(HKEY_CURRENT_USER, "Software\\Cisco Systems\\CCNMediaTerm\\1.0", "UseFixedTransmitPort", false);
	if (bUseFixedTransmitPort)
	{
		unsigned short localTransmitPort = (unsigned short)(GetRegKeyLong(HKEY_CURRENT_USER, "Software\\Cisco Systems\\CCNMediaTerm\\1.0", "TransmitPort", 0) % 65536);
		rtpSink->SetLocalPort(localTransmitPort);
	}
	resultCode = rtpSink->SetDestination(hostname, (short)nUDPPortNumber);
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "~SetAudioDestination : %d\n", resultCode);
	return resultCode;
}



STDMETHODIMP 
CCCNMediaTerm::SetAudioReceivePort(long nUDPPortNumber)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState())
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "SetAudioReceivePort %d\n", nUDPPortNumber);
	CSingleLock cLock(&Mutex);
	cLock.Lock();
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "SetAudioReceivePort : entered Mutex\n");
	bool restart = false;
	int result(0);
	int resultCode(0);

	if (brxRunning)
	{
		result = StopRX();
		if (result < 0)
		{
			resultCode = -10;
		}
		restart = true;
	}

	if (resultCode == 0)
	{
		result = rtpSource->SetReceivePort(nUDPPortNumber);
		if (result < 0)
		{
			resultCode = -20;
		}
	}

	if (resultCode == 0)
	{
		if (restart)
		{
			result = StartRX(waveoutDeviceID);
		}
		if (result < 0)
		{
			resultCode = -30;
		}
	}
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "~SetAudioReceivePort : %d\n", resultCode);
	return resultCode;
}

STDMETHODIMP 
CCCNMediaTerm::StartPlayingFileTX(BSTR Filename, unsigned long Mode, long * Cookie)
{
	USES_CONVERSION;
	std::string filename = std::string(W2A((wchar_t *)Filename));
	tracer.tracef(EE, "StartPlayingFileTX file = %s, mode = %d\n", filename.c_str(), Mode);
	CSingleLock cLock(&Mutex);
	cLock.Lock();
	tracer.tracef(ARB, "StartPlayingFileTX : calling RemoveNonPlayingFiles()\n");
	RemoveNonPlayingFiles();
	if (btxRunning)
	{
		tracer.tracef(ARB, "StartPlayingFileTX : brxRunning = true. Playing file\n");
		FilePlay *newFilePlay = new FilePlay(this, txPCMMixer, txDuration, true);
		newFilePlay->LoadFile(filename.c_str());
		if (Mode == 0)
		{
			newFilePlay->PlayInLoop();
		}
		newFilePlay->SetVolume(100);
		newFilePlay->SetDestinationFormat(WaveFormat::GetWaveFormat(WaveFormat_PCM_16_8_1));
		tracer.tracef(ARB, "StartPlayingFileTX : Entering filesPlayingMutex\n");
		EnterCriticalSection(&filesPlayingMutex);
		filesPlaying[newFilePlay->GetCookie()] = newFilePlay;
		*Cookie = newFilePlay->GetCookie();
		newFilePlay->Start();
		LeaveCriticalSection(&filesPlayingMutex);
		tracer.tracef(ARB, "StartPlayingFileTX : Left filesPlayingMutex\n");
	}
	tracer.tracef(EE, "~StartPlayingFileTX : cookie = %d\n", *Cookie);
	return S_OK;
}

STDMETHODIMP 
CCCNMediaTerm::StopPlayingFileTX(unsigned long Cookie)
{
	tracer.tracef(EE, "StopPlayingFileTX : cookie = %u\n", Cookie);
	CSingleLock cLock(&Mutex);
	cLock.Lock();
	tracer.tracef(ARB, "StopPlayingFileTX : Entering filesPlayingMutex\n");
	EnterCriticalSection(&filesPlayingMutex);
	map<unsigned int, FilePlay *>::iterator iter = filesPlaying.find(Cookie);
	if (iter != filesPlaying.end())
	{
		FilePlay *filePlay = (*iter).second;
		if (filePlay->Transmitting() == true)
		{
			filePlay->StopImmediately();
		}
	}
	LeaveCriticalSection(&filesPlayingMutex);
	tracer.tracef(ARB, "StopPlayingFileTX : Left filesPlayingMutex\n");
	tracer.tracef(EE, "~StopPlayingFileTX\n");
	return NOERROR;
}

STDMETHODIMP 
CCCNMediaTerm::StartPlayingFileRX(BSTR Filename, unsigned long Mode, unsigned long waveoutDeviceID, long * Cookie)
{
	USES_CONVERSION;
	std::string filename = std::string(W2A((wchar_t *)Filename));
	tracer.tracef(EE, "StartPlayingFileRX file = %s, mode = %d, waveOutDeviceID = %d\n", filename.c_str(), Mode, waveoutDeviceID);
	CSingleLock cLock(&Mutex);
	cLock.Lock();
	tracer.tracef(ARB, "StartPlayingFileRX : calling RemoveNonPlayingFiles()\n");
	RemoveNonPlayingFiles();
	if (brxRunning && waveoutDeviceID == this->waveoutDeviceID)
	{
		tracer.tracef(ARB, "StartPlayingFileRX : playing file in audio stream\n");
		FilePlay *newFilePlay = new FilePlay(this, rxPCMMixer, rxDuration, false);
		newFilePlay->LoadFile(filename.c_str());
		if (Mode == 0)
		{
			newFilePlay->PlayInLoop();
		}
		newFilePlay->SetVolume(100);
		newFilePlay->SetDestinationFormat(WaveFormat::GetWaveFormat(WaveFormat_PCM_16_8_1));
		tracer.tracef(ARB, "StartPlayingFileRX : Entering filesPlayingMutex\n");
		EnterCriticalSection(&filesPlayingMutex);
		filesPlaying[newFilePlay->GetCookie()] = newFilePlay;
		*Cookie = newFilePlay->GetCookie();
		newFilePlay->Start();
		LeaveCriticalSection(&filesPlayingMutex);
		tracer.tracef(ARB, "StartPlayingFileRX : Left filesPlayingMutex\n");
	}
	else
	{
		if (bFilePlayRunning)
		{
			tracer.tracef(ARB, "StartPlayingFileRX : bFilePlayRunning = true. Calling StopPlayingFileRX\n");
			StopPlayingFileRX(extraFilePlay->GetCookie());
		}
		if (extraFilePlay == NULL)
		{
			tracer.tracef(ARB, "StartPlayingFileRX : extraFilePlay = NULL. Playing file\n");
			extraFilePlay = new FilePlay(this, filePlayWaveSink, 60, false);
			extraFilePlay->LoadFile(filename.c_str());
			extraFilePlay->SetDestinationFormat(WaveFormat::GetWaveFormat(WaveFormat_PCM_16_8_1));
			extraFilePlay->SetVolume(100);
			if (Mode == 0)
			{
				extraFilePlay->PlayInLoop();
			}
			filePlayWaveSink->SetDeviceID(waveoutDeviceID);
			filePlayWaveSink->SetFrameSize(60 * 16);
			filePlayWaveSink->SetNumFrames(10);
			filePlayWaveSink->SetFormat(WaveFormat_PCM_16_8_1);
			tracer.tracef(ARB, "StartPlayingFileRX : Entering filesPlayingMutex\n");
			EnterCriticalSection(&filesPlayingMutex);
			*Cookie = extraFilePlay->GetCookie();
			LeaveCriticalSection(&filesPlayingMutex);
			tracer.tracef(ARB, "StartPlayingFileRX : Left filesPlayingMutex\n");
			extraFilePlay->Start();
			filePlayWaveSink->StartSinkThread();
			bFilePlayRunning = true;
		}
	}
	tracer.tracef(EE, "~StartPlayingFileRX : cookie = %u\n", *Cookie);
	return S_OK;
}


STDMETHODIMP 
CCCNMediaTerm::StopPlayingFileRX(unsigned long Cookie)
{
	tracer.tracef(EE, "StopPlayingFileRX %u\n", Cookie);
	CSingleLock cLock(&Mutex);
	cLock.Lock();
	tracer.tracef(ARB, "StopPlayingFileRX : Entering filesPlayingMutex\n");
	EnterCriticalSection(&filesPlayingMutex);
	map<unsigned int, FilePlay *>::iterator iter = filesPlaying.find(Cookie);
	if (iter != filesPlaying.end())
	{
		FilePlay *filePlay = (*iter).second;
		if (filePlay->Transmitting() == false)
		{
			filePlay->StopImmediately();
		}
		LeaveCriticalSection(&filesPlayingMutex);
		tracer.tracef(ARB, "StopPlayingFileRX : Left filesPlayingMutex\n");
	}
	else 
	{
		LeaveCriticalSection(&filesPlayingMutex);
		tracer.tracef(ARB, "StopPlayingFileRX : Left filesPlayingMutex\n");
		if (extraFilePlay)
		{
			tracer.tracef(DET, "StopPlayingFileRX : extraFilePlay->cookie = %d\n", extraFilePlay->GetCookie());
			if (Cookie == extraFilePlay->GetCookie())
			{
				tracer.tracef(DET, "StopPlayingFileRX : calling filePlayWaveSink->StopSinkThread\n");
				int i;
				i = filePlayWaveSink->StopSinkThread();
				if (i != 0)
				{
					tracer.tracef(ERR, "StopPlayingFileRX : e-filePlayWaveSink->StopSinkThread() : %d\n", i);
				}
				i = extraFilePlay->StopImmediately();
				if (i != 0)
				{
					tracer.tracef(ERR, "StopPlayingFileRX : e-extraFilePlay->StopImmediately() : %d\n", i);
				}
				if (!extraFilePlay->IsRunning())
				{
					bFilePlayRunning = false;
					tracer.tracef(DET, "StopPlayingFileRX : deleting extraFilePlay\n");
					delete extraFilePlay;
					extraFilePlay = NULL;
				}
				else
				{
					int i=10;
				}
			}
		}
	}
	tracer.tracef(EE, "~StopPlayingFileRX\n");
	return S_OK;
}

STDMETHODIMP 
CCCNMediaTerm::StartTX(unsigned long waveinDeviceID)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState())
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "StartTX\n");
	CSingleLock cLock(&Mutex);
	cLock.Lock();
	int resultCode(0);
	
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "StartTX : 0\n");
	if (!btxRunning)
	{
		// find out if pre-emphasis needs to be done, and if so with what parameters
		btxPreEmphasis = GetRegKeyBool(HKEY_CURRENT_USER, "Software\\Cisco Systems\\CCNMediaTerm\\1.0", "MicrophonePreprocess", true);
		long txFIRFilter = GetRegKeyLong(HKEY_CURRENT_USER, "Software\\Cisco Systems\\CCNMediaTerm\\1.0", "TxFIRFilter", 1);
		if (txFIRFilter == 2)
		{
			outputFIRTransformer->SetCoefficients(FIR_TELECASTER_TX, FIR_TELECASTER_TX_NUM, 16);
		}
		else
		{
			outputFIRTransformer->SetCoefficients(FIR_ITU_TX_IRS, FIR_ITU_TX_IRS_NUM, 16);
		}

		this->waveinDeviceID = waveinDeviceID;
		waveSource->SetDeviceID(waveinDeviceID);

		AudioSource *temp = waveSource;

		if (btxPreEmphasis)
		{
			Connect(temp, outputFIRTransformer);
			temp = outputFIRTransformer;
		}
		if (btxVAD)
		{
			Connect(temp, voiceActivityDetector);
			temp = voiceActivityDetector;
		}
		Connect(temp, txPCMMixer);
		Connect(txPCMMixer, txTransformer);
		Connect(txTransformer, rtpSink);

		do
		{
			int result(0);
			result = rtpSink->StartSink();
			if (result < 0)
			{
				resultCode = -10;
				break;
			}
			result = txTransformer->StartTransform();
			if (result < 0)
			{
				rtpSink->StopSink();
				resultCode = -20;
				break;
			}
			result = txPCMMixer->StartTransform();
			if (result < 0)
			{
				rtpSink->StopSink();
				txTransformer->StopTransform();
				resultCode = -30;
				break;
			}
			if (btxVAD)
			{
				voiceActivityDetector->EnableDetection();
				result = voiceActivityDetector->StartTransform();
				if (result < 0)
				{
					rtpSink->StopSink();
					txTransformer->StopTransform();
					txPCMMixer->StopTransform();
					resultCode = -40;
					break;
				}
			}
			if (btxPreEmphasis)
			{
				result = outputFIRTransformer->StartTransform();
				if (result < 0)
				{
					rtpSink->StopSink();
					txTransformer->StopTransform();
					txPCMMixer->StopTransform();
					if (btxVAD)
					{
						voiceActivityDetector->StopTransform();
					}
					resultCode = -50;
					break;
				}
			}
			result = waveSource->StartSourceThread();
			if (result < 0)
			{
				rtpSink->StopSink();
				txTransformer->StopTransform();
				txPCMMixer->StopTransform();
				if (btxVAD)
				{
					voiceActivityDetector->StopTransform();
				}
				if (btxPreEmphasis)
				{
					outputFIRTransformer->StopTransform();
				}
				resultCode = -60;
				break;
			}
		} while (false);
	}
	if (resultCode == 0)
	{
		btxRunning = true;
	}
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "~StartTX : %d\n", resultCode);
	return resultCode;
}





STDMETHODIMP 
CCCNMediaTerm::StopTX()
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState())
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "StopTX\n");
	CSingleLock cLock(&Mutex);
	cLock.Lock();

	bool bError(false);
	int resultCode(0);
	tracer.tracef(ARB, "StopTX : after cLock.Lock()\n");
	if (btxRunning)
	{
		int result(0);
		result = StopAllFiles(true);
		if (result < 0)
		{
			bError = true;
			tracer.tracef(ERR, "StopTX : e-StopAllFiles %d\n", result);
		}
		tracer.tracef(ARB, "StopTX : before waveSource->StopSourceThread()\n");
		result = waveSource->StopSourceThread();
		if (result < 0)
		{
			bError = true;
			tracer.tracef(ERR, "StopTX : e-waveSource->StopSourceThread %d\n", result);
		}
		if (btxPreEmphasis)
		{
			tracer.tracef(ARB, "StopTX : before outputFIRTransformer->StopTransform()\n");
			result = outputFIRTransformer->StopTransform();
			if (result < 0)
			{
				bError = true;
				tracer.tracef(ERR, "StopTX : e-outputFIRTransformer->StopTransform %d\n", result);
			}
		}
		if (btxVAD)
		{
			tracer.tracef(ARB, "StopTX : before voiceActivityDetector->StopTransform()\n");
			result = voiceActivityDetector->StopTransform();
			if (result < 0)
			{
				bError = true;
				tracer.tracef(ERR, "StopTX : e-voiceActivityDetector->StopTransform %d\n", result);
			}
		}
		tracer.tracef(ARB, "StopTX : before txPCMMixer->StopTransform()\n");
		result = txPCMMixer->StopTransform();
		if (result < 0)
		{
			bError = true;
			tracer.tracef(ERR, "StopTX : e-txPCMMixer->StopTransform %d\n", result);
		}
		tracer.tracef(ARB, "StopTX : before txTransformer->StopTransform()\n");
		result = txTransformer->StopTransform();
		if (result < 0)
		{
			bError = true;
			tracer.tracef(ERR, "StopTX : e-txTransformer->StopTransform %d\n", result);
		}
		tracer.tracef(ARB, "StopTX : before rtpSink->StopSink()\n");
		result = rtpSink->StopSink();
		if (result < 0)
		{
			bError = true;
			tracer.tracef(ERR, "StopTX : e-rtpSink->StopSink %d\n", result);
		}
		tracer.tracef(ARB, "StopTX : before Disconnects\n");

		AudioSource *temp = waveSource;
		if (btxPreEmphasis)
		{
			Disconnect(temp, outputFIRTransformer);
			temp = outputFIRTransformer;
		}
		if (btxVAD)
		{
			Disconnect(temp, voiceActivityDetector);
			temp = voiceActivityDetector;
		}
		Disconnect(temp, txPCMMixer);
		Disconnect(txPCMMixer, txTransformer);
		Disconnect(txTransformer, rtpSink);

		tracer.tracef(ARB, "StopTX : after Disconnects\n");
	}
	if (!bError)
	{
		btxRunning = false;
	}
	else
	{
		resultCode = -10;
	}
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "~StopTX : %d\n", resultCode);
	return resultCode;
}




STDMETHODIMP 
CCCNMediaTerm::StartRX(unsigned long waveoutDeviceID)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState())
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "StartRX\n");
	CSingleLock cLock(&Mutex);
	cLock.Lock();

	int result(0);
	int returnCode(0);

	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "StartRX : brxRunning = %d\n", brxRunning);
	if (!brxRunning)
	{
		brxPostEmphasis = GetRegKeyBool(HKEY_CURRENT_USER, "Software\\Cisco Systems\\CCNMediaTerm\\1.0", "SpeakerPostprocess", true);
		brxVolumeLimit = GetRegKeyBool(HKEY_CURRENT_USER, "Software\\Cisco Systems\\CCNMediaTerm\\1.0", "LimitVolume", true);

		// set up the post emphasis filter
		long rxFIRFilter = GetRegKeyLong(HKEY_CURRENT_USER, "Software\\Cisco Systems\\CCNMediaTerm\\1.0", "RxFIRFilter", 1);
		if (rxFIRFilter == 2)
		{
			inputFIRTransformer->SetCoefficients(FIR_TELECASTER_RX, FIR_TELECASTER_RX_NUM, 16);
		}
		else
		{
			inputFIRTransformer->SetCoefficients(FIR_ITU_RX_IRS, FIR_ITU_RX_IRS_NUM, 16);
		}

		// set up the limiter
		// set up RTP jitter buffer and fixed size audio buffer
		long jitterBufferDepth = GetRegKeyLong(HKEY_CURRENT_USER, "Software\\Cisco Systems\\CCNMediaTerm\\1.0", "JitterBufferTime", 100);
		rtpJitterBuffer->SetJitterBufferDepth(jitterBufferDepth);
		fixedSizeAudioBuffer->SetOutputDuration(rxDuration);
		int numFrames = _winmajor < 5 ? 10 : 5;
		bool  useDynamicJitterBuffer = GetRegKeyBool(HKEY_CURRENT_USER, "Software\\Cisco Systems\\CCNMediaTerm\\1.0", "UseDynamicJitterBuffer", true);
		if (useDynamicJitterBuffer)
		{
			waveSink->SetNumFrames(numFrames);
		}
		else
		{
			waveSink->SetNumFrames(max(numFrames, jitterBufferDepth/rxDuration));
		}

		double threshold;
		double lossIncrement;
		double lossDecrement;
		std::string strThreshold = GetRegKeyString(HKEY_CURRENT_USER, "Software\\Cisco Systems\\CCNMediaTerm\\1.0", "LimiterThreshold", "-8.0");
		std::string strLossIncrement = GetRegKeyString(HKEY_CURRENT_USER, "Software\\Cisco Systems\\CCNMediaTerm\\1.0", "LimiterLossIncrement", "0.075");
		std::string strLossDecrement = GetRegKeyString(HKEY_CURRENT_USER, "Software\\Cisco Systems\\CCNMediaTerm\\1.0", "LimiterLossDecrement", "-0.00075");
		threshold = atof(strThreshold.c_str());
		lossIncrement = atof(strLossIncrement.c_str());
		lossDecrement = atof(strLossDecrement.c_str());
		limiter->SetParameters(threshold, lossIncrement, lossDecrement);

		this->waveoutDeviceID = waveoutDeviceID;
		waveSink->SetDeviceID(waveoutDeviceID);

		Connect(rtpSource, rxTransformer);
		Connect(rxTransformer, rtpJitterBuffer);
		Connect(rtpJitterBuffer, fixedSizeAudioBuffer);
		Connect(fixedSizeAudioBuffer, rxPCMMixer);
		AudioSource *tempSource = rxPCMMixer;
		if (brxPostEmphasis)
		{
			Connect(tempSource, inputFIRTransformer);
			tempSource = inputFIRTransformer;
		}
		if (brxVolumeLimit)
		{
			Connect(tempSource, limiter);
			tempSource = limiter;
		}
		Connect(tempSource, waveSink);

		do
		{
			result = rxTransformer->StartTransform();
			if (result < 0)
			{
				returnCode = -10;
				break;
			}
			result = rtpJitterBuffer->StartTransform();
			if (result < 0)
			{
				rxTransformer->StopTransform();
				returnCode = -20;
				break;
			}
			result = fixedSizeAudioBuffer->StartTransform();
			if (result < 0)
			{
				rxTransformer->StopTransform();
				rtpJitterBuffer->StopTransform();
				returnCode = -30;
				break;
			}
			result = rxPCMMixer->StartTransform();
			if (result < 0)
			{
				rxTransformer->StopTransform();
				rtpJitterBuffer->StopTransform();
				rxTransformer->StopTransform();
				returnCode = -40;
				break;
			}
			if (brxPostEmphasis)
			{
				result = inputFIRTransformer->StartTransform();
				if (result < 0)
				{
					rxTransformer->StopTransform();
					rtpJitterBuffer->StopTransform();
					rxTransformer->StopTransform();
					rxPCMMixer->StopTransform();
					returnCode = -50;
					break;
				}
			}
			if (brxVolumeLimit)
			{
				result = limiter->StartTransform();
				if (result < 0)
				{
					rxTransformer->StopTransform();
					rtpJitterBuffer->StopTransform();
					rxTransformer->StopTransform();
					rxPCMMixer->StopTransform();
					inputFIRTransformer->StopTransform();
					returnCode = -60;
					break;
				}
			}
			result = waveSink->StartSinkThread();
			if (result < 0)
			{
				rxTransformer->StopTransform();
				rtpJitterBuffer->StopTransform();
				rxTransformer->StopTransform();
				rxPCMMixer->StopTransform();
				inputFIRTransformer->StopTransform();
				limiter->StopTransform();
				returnCode = -70;
				break;
			}
			result = rtpSource->StartSourceThread();
			if (result < 0)
			{
				waveSink->StopSinkThread();
				rxTransformer->StopTransform();
				rtpJitterBuffer->StopTransform();
				rxTransformer->StopTransform();
				rxPCMMixer->StopTransform();
				inputFIRTransformer->StopTransform();
				limiter->StopTransform();
				returnCode = -80;
				break;
			}
		} while (false);
	}			
	if (returnCode == 0)
	{
		brxRunning = true;
	}
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "~StartRX : returning %d\n", returnCode);
	return returnCode;
}

STDMETHODIMP 
CCCNMediaTerm::StopRX()
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState())
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "StopRX\n");
	CSingleLock cLock(&Mutex);
	cLock.Lock();

	bool bError = false;
	int resultCode(0);
	tracer.tracef(ARB, "StopRX : brxRunning = %d\n", brxRunning);
	if (brxRunning)
	{
		int result(0);
		tracer.tracef(ARB, "StopRX : calling StopAllFiles(false)\n");
		result = StopAllFiles(false);
		if (result < 0)
		{
			bError = true;
			tracer.tracef(ERR, "StopRX : e-StopAllFiles = %d\n", result);
		}
		tracer.tracef(ARB, "StopRX : calling waveSink->StopSinkThread()\n");
		result = waveSink->StopSinkThread();
		if (result < 0)
		{
			bError = true;
			tracer.tracef(ERR, "StopRX : e-waveSink->StopSinkThread = %d\n", result);
		}
		tracer.tracef(ARB, "StopRX : calling rtpSource->UnprepareSource()\n");
		result = rtpSource->UnprepareSource();
		if (result < 0)
		{
			bError = true;
			tracer.tracef(ERR, "StopRX : e-rtpSource->UnprepareSource = %d\n", result);
		}
		tracer.tracef(ARB, "StopRX : calling rtpSource->StopSourceThread()\n");
		result = rtpSource->StopSourceThread();
		if (result < 0)
		{
			bError = true;
			tracer.tracef(ERR, "StopRX : e-rtpSource->StopSourceThread = %d\n", result);
		}
		if (brxVolumeLimit)
		{
			tracer.tracef(ARB, "StopRX : calling limiter->StopTransform()\n");
			result = limiter->StopTransform();
			if (result < 0)
			{
				bError = true;
				tracer.tracef(ERR, "StopRX : e-limiter->StopTransform = %d\n", result);
			}
		}
		if (brxPostEmphasis)
		{
			tracer.tracef(ARB, "StopRX : calling inputFIRTransformer->StopTransform()\n");
			result = inputFIRTransformer->StopTransform();
			if (result < 0)
			{
				bError = true;
				tracer.tracef(ERR, "StopRX : e-inputFIRTransformer->StopTransform = %d\n", result);
			}
		}
		tracer.tracef(ARB, "StopRX : calling rxPCMMixer->StopTransform()\n");
		result = rxPCMMixer->StopTransform();
		if (result < 0)
		{
			bError = true;
			tracer.tracef(ERR, "StopRX : e-rxPCMMixer->StopTransform = %d\n", result);
		}
		tracer.tracef(ARB, "StopRX : calling rxTransformer->StopTransform()\n");
		result = rxTransformer->StopTransform();
		if (result < 0)
		{
			bError = true;
			tracer.tracef(ERR, "StopRX : e-rxTransformer->StopTransform = %d\n", result);
		}

		tracer.tracef(ARB, "StopRX : disconnecting filters\n");
		Disconnect(rtpSource, rxTransformer);
		Disconnect(rxTransformer, rtpJitterBuffer);
		Disconnect(rtpJitterBuffer, fixedSizeAudioBuffer);
		Disconnect(fixedSizeAudioBuffer, rxPCMMixer);
		AudioSource *tempSource = rxPCMMixer;
		if (brxPostEmphasis)
		{
			Disconnect(tempSource, inputFIRTransformer);
			tempSource = inputFIRTransformer;
		}
		if (brxVolumeLimit)
		{
			Disconnect(tempSource, limiter);
			tempSource = limiter;
		}
		Disconnect(tempSource, waveSink);
	}
	if (!bError)
	{
		brxRunning = false;
	}
	else
	{
		resultCode = -10;
	}
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "~StopRX : %d\n", resultCode);
	return resultCode;
}

int 
CCCNMediaTerm::RemoveNonPlayingFiles()
{
	tracer.tracef(EE, "RemoveNonPlayingFiles\n");
	tracer.tracef(ARB, "RemoveNonPlayingFiles : Entering filesPlayingMutex\n");
	EnterCriticalSection(&filesPlayingMutex);
	vector<map<unsigned int, FilePlay *>::iterator> filesToDelete;
	map<unsigned int, FilePlay *>::iterator iter = filesPlaying.begin();
	while (iter != filesPlaying.end())
	{
		FilePlay *filePlay = (*iter).second;
		tracer.tracef(DET, "RemoveNonPlayingFiles : file found with cookie = %d\n", filePlay->GetCookie());
		if (!(filePlay->IsRunning()))
		{
			tracer.tracef(DET, "RemoveNonPlayingFiles : file with cookie = %d is not playing\n", filePlay->GetCookie());
			filesToDelete.push_back(iter);
		}
		iter++;
	}
	tracer.tracef(DET, "RemoveNonPlayingFiles : deleting non-playing files from list\n");
	for (int i=0; i<filesToDelete.size(); i++)
	{
		map<unsigned int, FilePlay *>::iterator iter = filesToDelete[i];
		FilePlay *filePlay = (*iter).second;
		filesPlaying.erase(iter);
		delete filePlay;
	}
	if (extraFilePlay)
	{
		if (!extraFilePlay->IsRunning())
		{
			tracer.tracef(DET, "RemoveNonPlayingFiles : extrafileplay found with cookie = %d\n", extraFilePlay->GetCookie());
			tracer.tracef(DET, "RemoveNonPlayingFiles : deleting extraFilePlay\n");
			delete extraFilePlay;
			extraFilePlay = NULL;
		}
	}
	LeaveCriticalSection(&filesPlayingMutex);
	tracer.tracef(ARB, "RemoveNonPlayingFiles : Left filesPlayingMutex\n");
	tracer.tracef(EE, "~RemoveNonPlayingFiles\n");
	return NOERROR;
}

int
CCCNMediaTerm::StopAllFiles(bool transmitSide)
{
	tracer.tracef(EE, "StopAllFiles %d\n", transmitSide);
	tracer.tracef(ARB, "StopAllFiles : Entering filesPlayingMutex\n");
	EnterCriticalSection(&filesPlayingMutex);
	map<unsigned int, FilePlay *>::iterator iter = filesPlaying.begin();
	while(iter != filesPlaying.end())
	{
		FilePlay *filePlay = (*iter).second;
		tracer.tracef(DET, "StopAllFiles : file with cookie = %d\n", filePlay->GetCookie());
		if (filePlay->IsRunning() && filePlay->Transmitting() == transmitSide)
		{
			tracer.tracef(DET, "StopAllFiles : stopping filePlay with cookie = %d\n", filePlay->GetCookie());
			filePlay->StopImmediately();
		}
		iter++;
	}
	LeaveCriticalSection(&filesPlayingMutex);
	tracer.tracef(ARB, "StopAllFiles : Left filesPlayingMutex\n");
	tracer.tracef(EE, "~StopAllFiles\n");
	return 0;
}

STDMETHODIMP CCCNMediaTerm::SetSpeakerVolume(unsigned long deviceID, unsigned long volume)
{
	int result(0);
	if (deviceID == filePlayWaveSink->GetDeviceID() && bFilePlayRunning)
	{
		result = filePlayWaveSink->SetVolume(volume);
	}
	else if (deviceID == waveSink->GetDeviceID() && brxRunning)
	{
		result = waveSink->SetVolume(volume);
	}
	return result;
}



// sets the components internal setting for mic volume. max volume = system's current mic volume setting
STDMETHODIMP 
CCCNMediaTerm::SetMicrophoneVolume(unsigned long deviceID, unsigned long volume)
{
	waveSource->SetVolume(volume);
	return S_OK;
}



// changes the system's microphone volume setting
HRESULT 
CCCNMediaTerm::SetSystemMicrophoneVolume(unsigned long volume)
{
	return S_OK;
}


bool
CCCNMediaTerm::SpeakerAvailable() 
{
	bool available = true;
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "SpeakerAvailable\n");

	WAVEFORMATEX waveFormatex;
	waveFormatex.nAvgBytesPerSec = 16000;
	waveFormatex.wFormatTag = WAVE_FORMAT_PCM;
	waveFormatex.wBitsPerSample = 16;
	waveFormatex.nChannels = 1;
	waveFormatex.nBlockAlign = 2;
	waveFormatex.nSamplesPerSec = 8000;
	waveFormatex.cbSize = 0;

	// see if the speaker is available
	HWAVEOUT hWaveOut;
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "SpeakerAvailable : opening wave device\n");
	MMRESULT mmr = waveOutOpen(&hWaveOut, WAVE_MAPPER, &waveFormatex, NULL, NULL, CALLBACK_NULL);
	if (mmr != MMSYSERR_NOERROR) {
		available = false;
		int retryInterval = 2000;
		tracer.tracef(SDI_LEVEL_ERROR, "SpeakerAvailable : error : could not open wave device. waveOutOpen returned 0x%x\n", mmr);
		tracer.tracef(SDI_LEVEL_ARBITRARY, "SpeakerAvailable : trying again after %d ms\n", retryInterval);
		Sleep(retryInterval);
		MMRESULT mmr = waveOutOpen(&hWaveOut, WAVE_MAPPER, &waveFormatex, NULL, NULL, CALLBACK_NULL);
		if (mmr != MMSYSERR_NOERROR) {
			tracer.tracef(SDI_LEVEL_ERROR, "SpeakerAvailable : error : still could not open wave device. waveOutOpen returned 0x%x\n", mmr);
		}
		else {
			tracer.tracef(SDI_LEVEL_ARBITRARY, "SpeakerAvailable : got the speaker second time\n");
			waveOutReset(hWaveOut);
			waveOutClose(hWaveOut);
			available = true;
		}
	}
	else {
		waveOutReset(hWaveOut);
		waveOutClose(hWaveOut);
		available = true;
	}
	
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "~SpeakerAvailable : returning %d\n", available);
	return available;
}




bool
CCCNMediaTerm::MicrophoneAvailable()
{
	bool available = true;
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "MicrophoneAvailable\n");

	WAVEFORMATEX waveFormatex;
	waveFormatex.nAvgBytesPerSec = 16000;
	waveFormatex.wFormatTag = WAVE_FORMAT_PCM;
	waveFormatex.wBitsPerSample = 16;
	waveFormatex.nChannels = 1;
	waveFormatex.nBlockAlign = 2;
	waveFormatex.nSamplesPerSec = 8000;
	waveFormatex.cbSize = 0;

	// see if the microphone is available
	HWAVEIN hwaveIn;
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "MicrophoneAvailable : opening wavein device\n");
	MMRESULT mmr = waveInOpen(&hwaveIn, WAVE_MAPPER, &waveFormatex, NULL, NULL, CALLBACK_NULL);
	if (mmr != MMSYSERR_NOERROR) {
		tracer.tracef(SDI_LEVEL_ERROR, "MicrophoneAvailable : error : could not open wavein device. waveInOpen returned 0x%x\n", mmr);
		available = false;
	}

	else {
		waveInReset(hwaveIn);
		waveInClose(hwaveIn);
	}
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "~MicrophoneAvailable : returning %d\n", available);
	return available;
}



STDMETHODIMP CCCNMediaTerm::SetFilePlayVolume(unsigned long cookie, unsigned long volume)
{
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "SetFilePlayVolume cookie=%u volume=%u\n", cookie, volume);
	if (bFilePlayRunning && cookie == extraFilePlay->GetCookie())
	{
		tracer.tracef(ARB, "SetFilePlayVolume : Setting volume for extraFilePlay\n");
		extraFilePlay->SetVolume((int)volume);
	}
	else
	{
		tracer.tracef(ARB, "SetFilePlayVolume : Entering filesPlayingMutex\n");
		EnterCriticalSection(&filesPlayingMutex);
		map<unsigned int, FilePlay *>::iterator iter = filesPlaying.find(cookie);
		if (iter != filesPlaying.end())
		{
			FilePlay *filePlay = (*iter).second;
			filePlay->SetVolume((int)volume);
		}
		LeaveCriticalSection(&filesPlayingMutex);
		tracer.tracef(ARB, "SetFilePlayVolume : Left filesPlayingMutex\n");
	}
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "~SetFilePlayVolume\n");
	return S_OK;
}

void
CCCNMediaTerm::FireEndOfFileRX(unsigned long cookie)
{
	int resultCode(0);
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "FireEndOfFileRX cookie=%u\n", cookie);
	if (bFilePlayRunning && extraFilePlay)
	{
		if (cookie == extraFilePlay->GetCookie())
		{
			resultCode = filePlayWaveSink->StopSinkThread();
			if (resultCode != 0)
			{
				tracer.tracef(ERR, "FireEndOfFileRX : e-filePlayWaveSink->StopSinkThread() : %d\n", resultCode);
			}
			tracer.tracef(DET, "FireEndOfFileRX : setting bFilePlayRunning to false\n");
			bFilePlayRunning = false;
		}
	}
	Fire_EndOfFileEventRX(cookie);
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "~FireEndOfFileRX\n");
}

void
CCCNMediaTerm::FireEndOfFileTX(unsigned long cookie)
{
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "FireEndOfFileTX cookie=%u\n", cookie);
	Fire_EndOfFileEventTX(cookie);
	tracer.tracef(SDI_LEVEL_ENTRY_EXIT, "~FireEndOfFileTX\n");
}

STDMETHODIMP CCCNMediaTerm::NetworkMonitor(unsigned long, unsigned long)
{
	return S_OK;
}

STDMETHODIMP CCCNMediaTerm::StartDtmfTone(long, long, long)
{
	return S_OK;
}

STDMETHODIMP CCCNMediaTerm::StopDtmfTone(void)
{
	return S_OK;
}

STDMETHODIMP CCCNMediaTerm::StopAudioReceive(void)
{
	return S_OK;
}

STDMETHODIMP CCCNMediaTerm::StartAudioReceive(void)
{
	return S_OK;
}

STDMETHODIMP CCCNMediaTerm::StopMicrophone(void)
{
	return S_OK;
}

STDMETHODIMP CCCNMediaTerm::StartMicrophone(void)
{
	return S_OK;
}

FilePlay::FilePlay(CCCNMediaTerm *mediaTerm, AudioSink *audioSink, int outputDuration, bool bTransmit)
{
	char subFacilityName[100];
	sprintf(subFacilityName, "FilePlay:%x", this);
	tracer.SetSubFacilityName(subFacilityName);
	SetTraceLevel();
	this->mediaTerm = mediaTerm;
	this->bTransmit = bTransmit;
	waveFileSource = new WaveFileSource();
	waveFileSource->AddFilePlayListener(this);
	acmTransformer = new ACMTransformer();
	acmTransformer->SetOutputDuration(outputDuration);
	pcmVolumeTransformer = new PCMVolumeTransformer();
	this->audioSink = NULL;
	bRunning = false;

	::Connect(waveFileSource, acmTransformer);
	::Connect(acmTransformer, pcmVolumeTransformer);
	Connect(audioSink);

	cookie = ++nextCookie;
}

FilePlay::~FilePlay()
{
	StopImmediately();
	delete waveFileSource;
	delete acmTransformer;
	delete pcmVolumeTransformer;
}

int
FilePlay::SetTraceLevel() {
	long SystemMask = 0;
	GetRegKeyLong(HKEY_CURRENT_USER, "Software\\Cisco Systems\\MTC\\Tracing", "FilePlay", 0x100000);
	if ((SystemMask = GetRegKeyLong(HKEY_CURRENT_USER, "Software\\Cisco Systems\\MTC\\Tracing", "AllComponents", 0x0)) == 0)
		SystemMask = GetRegKeyLong(HKEY_CURRENT_USER, "Software\\Cisco Systems\\MTC\\Tracing", "FilePlay", 0x100000);
	tracer.SetSystemMask(SystemMask);
	return NOERROR;
}

int
FilePlay::LoadFile(const char *fileName)
{
	return waveFileSource->OpenWaveFile(fileName);
}

int
FilePlay::PlayOnce()
{
	return waveFileSource->PlayOnce();
}

int
FilePlay::PlayInLoop()
{
	return waveFileSource->PlayInLoop();
}

int
FilePlay::Start()
{
	int returnCode(0);
	if (!bRunning)
	{
		int result(0);
		WAVEFORMATEX wfx;
		do
		{
			result = waveFileSource->GetFormat(&wfx);
			if (result < 0)
			{
				returnCode = -10;
				break;
			}
			result = acmTransformer->SetSourceFormat(wfx);
			if (result < 0)
			{
				returnCode = -20;
				break;
			}
			result = pcmVolumeTransformer->StartTransform();
			if (result < 0)
			{
				returnCode = -30;
				break;
			}
			pcmVolumeTransformer->SetVolume(25);
			result = acmTransformer->StartTransform();
			if (result < 0)
			{
				pcmVolumeTransformer->StopTransform();
				returnCode = -40;
				break;
			}
			waveFileSource->StartSourceThread();
			if (result < 0)
			{
				pcmVolumeTransformer->StopTransform();
				acmTransformer->StopTransform();
				returnCode = -50;
				break;
			}
		} while (false);
	}
	if (returnCode == 0)
	{
		bRunning = true;
	}
	return returnCode;
}

void FilePlay::Cleanup()
{	
	::Disconnect(waveFileSource, acmTransformer);
	acmTransformer->StopTransform();
	pcmVolumeTransformer->StopTransform();
	::Disconnect(acmTransformer, pcmVolumeTransformer);
	Disconnect();
	if (bTransmit)
	{
		mediaTerm->FireEndOfFileTX(cookie);
	}
	else
	{
		mediaTerm->FireEndOfFileRX(cookie);
	}
}

void FilePlay::CleanupImmediately()
{
	acmTransformer->StopTransform();
	pcmVolumeTransformer->StopTransform();
	::Disconnect(waveFileSource, acmTransformer);
	::Disconnect(acmTransformer, pcmVolumeTransformer);
	Disconnect();
	if (bTransmit)
	{
		mediaTerm->FireEndOfFileTX(cookie);
	}
	else
	{
		mediaTerm->FireEndOfFileRX(cookie);
	}
}

int
FilePlay::Stop()
{
	if (bRunning)
	{
		bRunning = false;
		waveFileSource->StopSourceThread();
		Cleanup();
	}
	
	return 0;
}

int
FilePlay::StopImmediately()
{
	if (bRunning)
	{
		bRunning = false;
		acmTransformer->StopTransform();
		waveFileSource->StopSourceThread();
		CleanupImmediately();
	}
	
	return 0;
}

bool
FilePlay::IsRunning()
{
	return bRunning;
}

int
FilePlay::SetVolume(int volume)
{

	return pcmVolumeTransformer->SetVolume(volume);
}

int
FilePlay::GetVolume()
{
	return pcmVolumeTransformer->GetVolume();
}

unsigned int
FilePlay::GetCookie()
{
	return cookie;
}

int
FilePlay::Connect(AudioSink *audioSink)
{
	Disconnect();
	if (audioSink)
	{
		::Connect(pcmVolumeTransformer, audioSink);
	}
	this->audioSink = audioSink;
	return 0;
}

int
FilePlay::Disconnect()
{
	if (audioSink)
	{
		::Disconnect(pcmVolumeTransformer, audioSink);
	}
	audioSink = NULL;
	return 0;
}

void
FilePlay::FilePlayStarted(WaveFileSource *waveFileSource, char *waveFileName)
{
	
}

void
FilePlay::FilePlayStopped(WaveFileSource *waveFileSource, char *waveFileName)
{
	if (bRunning)
	{
		bRunning = false;
		Cleanup();
	}
}

int
FilePlay::SetOutputDuration(int outputDuration)
{
	return 	acmTransformer->SetOutputDuration(outputDuration);
}

int
FilePlay::SetDestinationFormat(WAVEFORMATEX wfx)
{
	return acmTransformer->SetDestinationFormat(wfx);
}

bool
FilePlay::Transmitting()
{
	return bTransmit;
}